/****************************************************************************\
*
* magnetiX - Magnetic Scrolls Interpreter - Mac OS X port
*
* Written by Jan-Sebastian Schliemann <magnetiX@maczentrisch.de>
*
*
* the original Magnetic was:
*
* Written by Niclas Karlsson <nkarlsso@abo.fi>,
*            David Kinder <davidk.kinder@virgin.net>,
*            Stefan Meier <Stefan.Meier@if-legends.org> and
*            Paul David Doherty <pdd@if-legends.org>
*
* Copyright (C) 1997-2008  Niclas Karlsson
*
*
*
*     This program is free software; you can redistribute it and/or modify
*     it under the terms of the GNU General Public License as published by
*     the Free Software Foundation; either version 2 of the License, or
*     (at your option) any later version.
*
*     This program is distributed in the hope that it will be useful,
*     but WITHOUT ANY WARRANTY; without even the implied warranty of
*     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*     GNU General Public License for more details.
*
*     You should have received a copy of the GNU General Public License
*     along with this program; if not, write to the Free Software
*     Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.
*
\*****************************************************************************/

#import "GSBackgroundView.h"

@implementation GSBackgroundView


- (void)setupBackgroundLayer
{
	NSRect newTargetFrame = NSZeroRect;
	
	for (NSScreen *screen in NSScreen.screens) {
		
		NSSize size = screen.frame.size;
		if (size.width > newTargetFrame.size.width) {
			newTargetFrame.size.width = size.width;
		}
		if (size.height > newTargetFrame.size.height) {
			newTargetFrame.size.height = size.height;
		}
	}

	BOOL needsRot = NO;
	if (newTargetFrame.size.height > newTargetFrame.size.width) {
		needsRot = YES;
		CGFloat temp = newTargetFrame.size.width;
		newTargetFrame.size.width = newTargetFrame.size.height;
		newTargetFrame.size.height = temp;
	}
	
	
	if (NSEqualSizes(targetFrame.size, newTargetFrame.size)) {
		[self repositionBackgroundLayer];
		return;
	}
	
	
	targetFrame = newTargetFrame;
	
	[_backgroundLayer removeFromSuperlayer];
		
	CALayer *layer = [CALayer layer];
	layer.frame = newTargetFrame;
	layer.contents = [NSImage imageNamed:@"cleanedMetal"]; // based on free texture from: http://www.texturex.com/Metal-Textures/TextureX+steel+texture+shiney+piece+metal+silver+camo+pattern+Texture.jpg.php
	layer.zPosition = 0;
	layer.opacity = 0.3;
	
	if (needsRot) {
		layer.transform = CATransform3DMakeRotation(270.0 / 180.0 * M_PI, 0.0, 0.0, 1.0);
	}
	
	[self.layer insertSublayer:layer atIndex:0];
	self.backgroundLayer = layer;
	
	[self repositionBackgroundLayer];
	
	[self registerForDraggedTypes:@[NSFilenamesPboardType]];
}


- (NSDragOperation)draggingEntered:(id <NSDraggingInfo>)sender
{
	if ( [sender.draggingPasteboard.types containsObject:NSFilenamesPboardType] ) {
		return NSDragOperationCopy;
	}
	
	return NSDragOperationNone;
}


- (NSDragOperation)draggingUpdated:(id <NSDraggingInfo>)sender
{
	return NSDragOperationCopy;
}


- (BOOL)prepareForDragOperation:(id <NSDraggingInfo>)sender
{
	return YES;
}


- (BOOL)performDragOperation:(id <NSDraggingInfo>)sender
{
	NSPasteboard *pboard = sender.draggingPasteboard;
 
	if ([pboard.types containsObject:NSFilenamesPboardType]) {
		[[NSNotificationCenter defaultCenter] postNotificationName:@"droppedGameFiles" object:[pboard propertyListForType:NSFilenamesPboardType]];
	}
	
	return YES;
}


- (NSSize)resize:(NSSize)source to:(NSSize)target
{
	CGFloat sourceWidth  = source.width;
	CGFloat sourceHeight = source.height;
	
	CGFloat targetWidth  = target.width;
	CGFloat targetHeight = target.height;
	
	CGFloat scaleFactor  = 0.0;
	CGFloat scaledWidth  = targetWidth;
	CGFloat scaledHeight = targetHeight;
	
	if (!NSEqualSizes(source, target)) {
		
		CGFloat widthFactor  = targetWidth / sourceWidth;
		CGFloat heightFactor = targetHeight / sourceHeight;
		
		if (widthFactor > heightFactor) {
			scaleFactor = widthFactor;
		} else {
			scaleFactor = heightFactor;
		}
		
		scaledWidth  = round(sourceWidth  * scaleFactor);
		scaledHeight = round(sourceHeight * scaleFactor);
	}
	
	return NSMakeSize(scaledWidth, scaledHeight);
}


- (void)repositionBackgroundLayer
{
	if (!_backgroundLayer || !self.window) { return; }
	
	NSRect layerFrame = _backgroundLayer.frame;
	NSRect windowFrame = self.window.frame;
		
	if (windowFrame.size.height > targetFrame.size.height || windowFrame.size.width > targetFrame.size.width) {
		layerFrame.size = [self resize:targetFrame.size to:windowFrame.size];
	} else {
		layerFrame.size = targetFrame.size;
	}
	
	layerFrame.origin.x = - ceil((layerFrame.size.width - windowFrame.size.width) / 2);
	layerFrame.origin.y = - ceil((layerFrame.size.height - windowFrame.size.height) / 2);
	
	[NSAnimationContext beginGrouping];
	[[NSAnimationContext currentContext] setDuration:0.0];
	_backgroundLayer.frame = layerFrame;
	[NSAnimationContext endGrouping];
}


- (void)drawRect:(NSRect)dirtyRect
{
    [super drawRect:dirtyRect];
    
	[[NSColor colorWithWhite:0.0 alpha:1] set];
	NSRectFill(dirtyRect);
}


- (void)mouseDragged:(NSEvent *)event
{
	[self moveWindow:self.window withEvent:event];	
	[super mouseDragged:event];
}


- (void)mouseDown:(NSEvent *)event
{
	if (self.mousedOverView) {
		self.isMovingAllowed = NO;
		[[NSNotificationCenter defaultCenter] postNotificationName:@"clickedInSelectionWindow" object:nil];

	} else {
		self.isMovingAllowed = YES;

	}
	
	[super mouseDown:event];
}


- (void)moveWindow:(id)window withEvent:(NSEvent *)theEvent
{
	if (!_isMovingAllowed) { return; }
	
	NSOperatingSystemVersion version = [[NSProcessInfo processInfo] operatingSystemVersion];
	if ((version.majorVersion == 10 && version.minorVersion >= 11) || version.majorVersion > 10) {
		
		[window performWindowDragWithEvent:theEvent];
		return;
		
	}
	
	NSPoint where = [theEvent locationInWindow];
	
	NSRect rect = NSMakeRect(1, 1, 1, 1);
	rect.origin = where;
	
	where =  [window convertRectToScreen:rect].origin;
	NSPoint origin = [window frame].origin;
	
	while ((theEvent = [NSApp nextEventMatchingMask:NSEventMaskLeftMouseDown|NSEventMaskLeftMouseDragged|NSEventMaskLeftMouseUp untilDate:[NSDate distantFuture] inMode:NSEventTrackingRunLoopMode dequeue:YES]) && ([theEvent type]!=NSEventTypeLeftMouseUp)) {
		NSAutoreleasePool  *pool = [[NSAutoreleasePool alloc] init];
		
		NSRect rect = NSMakeRect(1, 1, 1, 1);
		rect.origin = [theEvent locationInWindow];
		
		NSPoint now = [window convertRectToScreen:rect].origin;
		origin.x += now.x-where.x;
		origin.y += now.y-where.y;
		
		[window setFrameOrigin:origin];
		where = now;
		[pool release];
	}
}


- (void)dealloc
{
	self.backgroundLayer = nil;
	self.mousedOverView = nil;
	[super dealloc];
}


@end
