/****************************************************************************\
*
* magnetiX - Magnetic Scrolls Interpreter - Mac OS X port
*
* Written by Jan-Sebastian Schliemann <magnetiX@maczentrisch.de>
*
*
* the original Magnetic was:
*
* Written by Niclas Karlsson <nkarlsso@abo.fi>,
*            David Kinder <davidk.kinder@virgin.net>,
*            Stefan Meier <Stefan.Meier@if-legends.org> and
*            Paul David Doherty <pdd@if-legends.org>
*
* Copyright (C) 1997-2008  Niclas Karlsson
*
*
*
*     This program is free software; you can redistribute it and/or modify
*     it under the terms of the GNU General Public License as published by
*     the Free Software Foundation; either version 2 of the License, or
*     (at your option) any later version.
*
*     This program is distributed in the hope that it will be useful,
*     but WITHOUT ANY WARRANTY; without even the implied warranty of
*     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*     GNU General Public License for more details.
*
*     You should have received a copy of the GNU General Public License
*     along with this program; if not, write to the Free Software
*     Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.
*
\*****************************************************************************/

#import <Cocoa/Cocoa.h>
#import <AVFoundation/AVFoundation.h>

@class MXCoverView;

@interface GameSelectController : NSObject
{
	IBOutlet id windowMenu;
	IBOutlet id magneticController;

	IBOutlet MXCoverView *mainCoverView;
	IBOutlet MXCoverView *mainCoverViewOld;
	IBOutlet MXCoverView *mainCoverViewOld_WA;
	IBOutlet MXCoverView *animCoverView;
	
	MXCoverView *mainCoverViewOldCurrent;

	IBOutlet NSWindow *theMainWindow;

	IBOutlet id selectSuperView;
	IBOutlet id gamesSuperView;

	IBOutlet NSVisualEffectView *topBorder;
	IBOutlet NSVisualEffectView *bottomBorder;
	
	IBOutlet NSButton *startButton;
	IBOutlet NSButton *musicButton;
	IBOutlet NSButton *locateGameFilesButton;
	IBOutlet NSTextView *locateGameFilesTextView;

	IBOutlet NSProgressIndicator *importPI;
	IBOutlet NSView *importPIContainer;
	
	NSRect newDeactivatedFrame;
	NSRect oldDeactivatedFrame;
	NSInteger oldDeactivatedIndex;
}

- (IBAction)toggleMusic:(id)sender;
- (IBAction)start:(id)sender;
- (void)stop;
- (IBAction)openHelpUrl:(id)sender;
- (IBAction)locateGameFiles:(id)sender;
- (NSString *)checkForFile:(NSString *)filename;
- (void)externalStart:(NSString *)filename;
- (void)playMidiData:(NSData *)midiData;
- (void)stopAllSounds;
- (void)changeGameSelectionWithNavigation:(unichar)character;
- (void)changeGameSelectionWithDirectJump:(unichar)character;
- (id)currentlyVisibleGame;

- (id)magneticController;
- (BOOL)startInstallationOf:(NSArray *)paths;
- (void)refreshAfterInstall;

@property (retain) AVAudioPlayer *titleMusicPlayer;
@property (retain) AVMIDIPlayer *wonderlandMidiPlayer;
@property (retain) NSTimer *autoPlayDelayer;

@property (retain) NSData *currentlyPlayingMidiData;

@property (retain) NSString *altInstallPath;
@property (retain) NSString *sessionPath;

@property (retain) NSArray *availableGames;
@property (retain) NSDictionary *availableGamesData;

@property (copy) NSString *currentGame;
@property (copy) NSString *lastGame;

@property (retain) NSArray *smallCoverViews;
@property (retain) NSImage *coverCollection;
@property (retain) NSImage *coverCollection_back;

@property (assign) NSInteger coverAnimType;

@property (assign) BOOL showBackCover;
@property (assign) BOOL isBackCoverToggle;
@property (assign) BOOL isCollectionToggle;
@property (assign) BOOL isLooping;
@property (assign) BOOL isSetup;

@property (retain) NSDate *lastCollectionClickTurn;

@end
