/****************************************************************************\
*
* magnetiX - Magnetic Scrolls Interpreter - Mac OS X port
*
* Written by Jan-Sebastian Schliemann <magnetiX@maczentrisch.de>
*
*
* the original Magnetic was:
*
* Written by Niclas Karlsson <nkarlsso@abo.fi>,
*            David Kinder <davidk.kinder@virgin.net>,
*            Stefan Meier <Stefan.Meier@if-legends.org> and
*            Paul David Doherty <pdd@if-legends.org>
*
* Copyright (C) 1997-2008  Niclas Karlsson
*
*
*
*     This program is free software; you can redistribute it and/or modify
*     it under the terms of the GNU General Public License as published by
*     the Free Software Foundation; either version 2 of the License, or
*     (at your option) any later version.
*
*     This program is distributed in the hope that it will be useful,
*     but WITHOUT ANY WARRANTY; without even the implied warranty of
*     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*     GNU General Public License for more details.
*
*     You should have received a copy of the GNU General Public License
*     along with this program; if not, write to the Free Software
*     Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.
*
\*****************************************************************************/

#import "GameSelectController.h"
#import "MXStatusView.h"
#import "MagneticController.h"
#import "MXCoverView.h"
#import "GSBackgroundView.h"
#import "WA_4SonomaAnimation.h"
#import <QuartzCore/QuartzCore.h>

@implementation GameSelectController


- (NSInteger)installFiles:(NSArray *)paths
{
	if (_availableGames.count == 10) { return 0; }
	
	NSDictionary *games = @{
		
		@"pawn" : @[@"gfx", @"mag", @"mp3"],
		@"guild" : @[@"gfx", @"mag"],
		@"cguild" : @[@"gfx", @"mag", @"hnt"],
		@"jinxter" : @[@"gfx", @"mag", @"mp3"],
		@"corrupt" : @[@"gfx", @"mag", @"mp3"],
		@"ccorrupt" : @[@"gfx", @"mag", @"hnt"],
		@"fish" : @[@"gfx", @"mag", @"mp3"],
		@"cfish" : @[@"gfx", @"mag", @"hnt"],
		@"myth" : @[@"gfx", @"mag"],
		@"wonder" : @[@"gfx", @"mag", @"hnt", @"mp3", @"snd"]

	};
	
	NSArray *zips = @[
		
		@"pawnall",
		@"guildall",
		@"guildmwall",
		@"jinxterall",
		@"corruptall",
		@"corrmwall",
		@"fishall",
		@"fishmwall",
		@"mythall",
		@"wonderall",
		@"corruptg",
		@"corruptm",
		@"corrupts",
		@"collecth",
		@"corrmwg",
		@"corrmwm",
		@"fishg",
		@"fishm",
		@"fishs",
		@"fishmwg",
		@"fishmwm",
		@"guildg",
		@"guildm",
		@"guildmwg",
		@"guildmwm",
		@"jinxterg",
		@"jinxterm",
		@"jinxters",
		@"mythg",
		@"mythm",
		@"pawng",
		@"pawnm",
		@"pawns",
		@"wonderg",
		@"wonderh",
		@"wonderm",
		@"wonders"

	];
	
	/* TODO: support ALT audios?
	 corrupts_alt.zip
	 corrupts_hq.zip
	 fishs_hq.zip
	 jinxters_hq.zip
	 pawns_hq.zip
	 */
	
	NSString *appSupport = self.applicationSupportFolder;
	NSFileManager *fileManager = NSFileManager.defaultManager;
	
	NSInteger count = 0;
	
	for (NSString *path in paths) {
		
		NSString *filename = path.lastPathComponent.lowercaseString;
		NSString *base = [filename stringByDeletingPathExtension];
		NSString *extension = [filename pathExtension];
		
		BOOL isDir = NO;
		
		if (!extension.length && [fileManager fileExistsAtPath:path isDirectory:&isDir] && isDir) {
			
			NSArray *files = [fileManager contentsOfDirectoryAtPath:path error:nil];
			if (files.count) {
				
				NSMutableArray *filesFull = NSMutableArray.array;
				for (NSString *file in files) {
					[filesFull addObject:[path stringByAppendingPathComponent:file]];
				}
				
				count += [self installFiles:filesFull];
				
			}
			
		} else if ([extension isEqualToString:@"zip"]) {
			
			NSDictionary *fileAttributes = [NSFileManager.defaultManager attributesOfItemAtPath:path error:nil];
			
			if ( ! [zips containsObject:base]) {
				
				NSArray *knownFileSizes = @[@(576447), @(1003897), @(683113), @(402085), @(363524), @(435291), @(1010388), @(103119), @(977594), @(2386395), @(191797), @(129782), @(611005), @(13473), @(428253), @(134772), @(175964), @(127578), @(281738), @(13473), @(256363), @(132306), @(195610), @(99213), @(13473), @(317314), @(104559), @(253471), @(122738), @(551206), @(30086), @(73058), @(270886), @(79487), @(562104), @(989948), @(8186), @(134130), @(1219476)];
				
				if ( ! [knownFileSizes containsObject:fileAttributes[NSFileSize]]) {
					continue;
				}
				
			} else {
				
				if ([fileAttributes[@"NSFileSize"] unsignedLongLongValue] > 5000000) {
					continue;
				}
				
			}
			
			NSString *destination = [appSupport stringByAppendingPathComponent:@"/temp-unzip"];
			if ( ! [fileManager createDirectoryAtPath:destination withIntermediateDirectories:NO attributes:nil error:nil]) {
				continue;
			}
			
			NSTask *unzip = [[NSTask alloc] init];
			unzip.executableURL = [NSURL fileURLWithPath:@"/usr/bin/unzip"];
			unzip.arguments = @[@"-u", @"-d", destination, path];
			
			[unzip launch];
			[unzip waitUntilExit];
			[unzip release];
			
			count += [self installFiles:@[destination]];
			
			[fileManager removeItemAtPath:destination error:nil];
			
		} else {
			
			NSArray *game = games[base];
			if ([game containsObject:extension] && ![self checkForFile:filename]) {
				
				NSString *targetPath = [appSupport stringByAppendingPathComponent:[self pathWithPotentialSubfolder:filename]];
				
				NSString *folder = [targetPath stringByDeletingLastPathComponent];
				if ( ! [fileManager fileExistsAtPath:folder]) {
					if ( ! [fileManager createDirectoryAtPath:folder withIntermediateDirectories:NO attributes:nil error:nil]) {
						continue;
					}
				}
				
				count += [fileManager copyItemAtPath:path toPath:targetPath error:nil];
				
			}
			
		}
		
	}
	
	return count;
}


- (void)refreshAfterInstall
{
	[self refreshCoverAnimType:0];
	
	NSArray *oldState = [_availableGames copy];
	
	[self checkAvailableGames];
	
	if (!_availableGames.count) {
		[oldState release];
		return;
	}
	
	if ( ! _isSetup) {
		if (oldState.count && oldState.count < _availableGames.count) {
			for (NSString *game in _availableGames) {
				if ( ! [oldState containsObject:game]) {
					[NSUserDefaults.standardUserDefaults setObject:game forKey:@"selectedGame"];
					break;
				}
			}
		}
		[oldState release];
		return;
	}
	
	if ( ! oldState.count) {
		[self finishSetup];
		
	} else {
		for (MXCoverView *cv in self.smallCoverViews) {
			[cv removeFromSuperview];
		}
		
		[self setupSmallCovers];
		[self setMainCover];
		[self updateCoverViews];
	}
	
	[theMainWindow makeKeyAndOrderFront:self];
	
	if (oldState.count && oldState.count < _availableGames.count && ! [magneticController selectedGame]) {
		for (NSString *game in _availableGames) {
			if ( ! [oldState containsObject:game]) {
				[[NSNotificationCenter defaultCenter] postNotificationName:@"updateCoverAnimType" object:@(2)];
				[self selectGameIfAvailable:game];
				break;
			}
		}
	}
	
	[oldState release];
}


- (NSString *)applicationSupportFolder
{
	NSArray *paths = NSSearchPathForDirectoriesInDomains(NSApplicationSupportDirectory, NSUserDomainMask, YES);
	NSString *basePath = ([paths count] > 0) ? paths[0] : NSTemporaryDirectory();
	
	NSString *appSuppPath = [basePath stringByAppendingPathComponent:@"magnetiX"];
	
	NSFileManager *fileManager = [NSFileManager defaultManager];
	if ( ![fileManager fileExistsAtPath:appSuppPath isDirectory:NULL] ) {
		[fileManager createDirectoryAtPath:appSuppPath withIntermediateDirectories:NO attributes:nil error:nil];
	}
	
	return appSuppPath;
}


- (NSString *)pathWithPotentialSubfolder:(NSString *)filename
{
	NSString *game = [filename stringByDeletingPathExtension];
	if ([game isEqualToString:@"cguild"]) { game = @"guildmw"; }
	else if ([game isEqualToString:@"ccorrupt"]) { game = @"corrmw"; }
	else if ([game isEqualToString:@"cfish"]) { game = @"fishmw"; }

	return [NSString stringWithFormat:@"%@all/%@", game, filename];
}


- (NSString *)checkForFile:(NSString *)filename
{
	NSString *path = [self checkForFile2:filename];
	if (path.length) { return path; }
	
	return [self checkForFile2:[self pathWithPotentialSubfolder:filename]];
}


- (NSString *)checkForFile2:(NSString *)filename
{
	NSFileManager *defaultManager = [NSFileManager defaultManager];
	NSString *fullPath = nil;
	
	if (_sessionPath) {
		
		fullPath = [_sessionPath stringByAppendingPathComponent:filename];
		
		if ([defaultManager fileExistsAtPath:fullPath]) {
			return fullPath;
		}
		
		return nil;
	}
	
	NSString *location = [@"~/Library/Application Support/magnetiX/" stringByExpandingTildeInPath];
	fullPath = [location stringByAppendingPathComponent:filename];
	
	if ([defaultManager fileExistsAtPath:fullPath]) {
		self.sessionPath = location;
		return fullPath;
	}
	
	
	location = @"/Library/Application Support/magnetiX/";
	fullPath = [location stringByAppendingPathComponent:filename];
	
	if ([defaultManager fileExistsAtPath:fullPath]) {
		self.sessionPath = location;
		return fullPath;
	}
	
		
	BOOL isDir = NO;
	
	if (![defaultManager fileExistsAtPath: _altInstallPath isDirectory:&isDir] || !isDir) {
		[self updateFileFolderPathFromBookmark];
	}
	
	NSString *userTrash = [@"~/.Trash/" stringByExpandingTildeInPath];
	
	if ([defaultManager fileExistsAtPath: [_altInstallPath stringByAppendingPathComponent:filename]]) {
		NSRange trashRange = [_altInstallPath rangeOfString:userTrash];
		NSRange trashesRange = [_altInstallPath rangeOfString:@"/.Trashes/"];
		
		if (trashRange.location != NSNotFound || trashesRange.location != NSNotFound) {
			return nil;
			
		} else {
			self.sessionPath = _altInstallPath;
			return [_altInstallPath stringByAppendingPathComponent:filename];
			
		}
	}
	
	return nil;
}


- (void)updateFileFolderPathFromBookmark
{
	NSData *bookmarkData = [[NSUserDefaults standardUserDefaults] dataForKey:@"gameFilesBookmarkData"];
	if (bookmarkData) {
		NSURL *URL = [NSURL URLByResolvingBookmarkData:bookmarkData options:NSURLBookmarkResolutionWithoutUI relativeToURL:nil bookmarkDataIsStale:nil error:nil];
		if (URL) {
			self.altInstallPath = [URL path];
		}
	}
}


- (void)setup
{
	self.isSetup = YES;
	
	theMainWindow.titleVisibility = NSWindowTitleHidden;
	theMainWindow.styleMask |= NSWindowStyleMaskFullSizeContentView;
	theMainWindow.titlebarAppearsTransparent = YES;
	if (@available(macOS 10.14.0, *)) {
		theMainWindow.appearance = [NSAppearance appearanceNamed:NSAppearanceNameDarkAqua];
	} else {
		theMainWindow.appearance = [NSAppearance appearanceNamed:NSAppearanceNameAqua];
	}
	
	[theMainWindow makeFirstResponder:nil];
	
	if (@available(macOS 11.0.0, *)) {
		NSRect frame = topBorder.frame;
		frame.size.height += 5.0;
		frame.origin.y -= 5.0;
		topBorder.frame = frame;
	}
	
	CALayer *topBorderLayer = topBorder.layer;
	topBorderLayer.zPosition = 20;
	topBorderLayer.backgroundColor = [[NSColor colorWithWhite:0.25 alpha:1] CGColor];
	topBorderLayer.borderColor = [[NSColor colorWithWhite:0.0 alpha:0.6] CGColor];
	topBorderLayer.borderWidth = 1;
	topBorderLayer.opacity = 0.5;
	topBorder.blendingMode = NSVisualEffectBlendingModeWithinWindow;
	if (@available(macOS 10.14.0, *)) {
		topBorder.material = NSVisualEffectMaterialTitlebar;
	}
	topBorder.appearance = [NSAppearance appearanceNamed:NSAppearanceNameVibrantDark];
	
	CALayer *bottomBorderLayer = bottomBorder.layer;
	bottomBorderLayer.zPosition = 20;
	bottomBorderLayer.backgroundColor = [[NSColor colorWithWhite:0.25 alpha:1] CGColor];
	bottomBorderLayer.borderColor = [[NSColor colorWithWhite:0.4 alpha:0.6] CGColor];
	bottomBorderLayer.borderWidth = 1;
	bottomBorderLayer.opacity = 0.5;
	bottomBorder.blendingMode = NSVisualEffectBlendingModeWithinWindow;
	if (@available(macOS 10.14.0, *)) {
		bottomBorder.material = NSVisualEffectMaterialTitlebar;
	}
	bottomBorder.appearance = [NSAppearance appearanceNamed:NSAppearanceNameVibrantDark];
	
	NSNotificationCenter *nc = [NSNotificationCenter defaultCenter];
	[nc addObserver:self selector:@selector(clickedInSelectionWindow:) name:@"clickedInSelectionWindow" object:nil];
	[nc addObserver:self selector:@selector(doubleClickedMainCover:) name:@"doubleClickedMainCover" object:nil];
	[nc addObserver:self selector:@selector(clickedMainCoverInArea:) name:@"clickedMainCoverInArea" object:nil];
	[nc addObserver:self selector:@selector(clickedArrow:) name:@"clickedArrow" object:nil];
	[nc addObserver:self selector:@selector(turnCover) name:@"clickedTurn" object:nil];
	[nc addObserver:self selector:@selector(updateUIStyle) name:@"updateUIStyle" object:nil];
	[nc addObserver:self selector:@selector(deactivateAllCovers) name:@"deactivateAllCovers" object:nil];
	[nc addObserver:self selector:@selector(updateCoverAnimType:) name:@"updateCoverAnimType" object:nil];
	[nc addObserver:self selector:@selector(processGamesFiles:) name:@"droppedGameFiles" object:nil];

	[self updateFileFolderPathFromBookmark];
	
	[selectSuperView setupBackgroundLayer];
	
	importPIContainer.hidden = YES;
	
	[self checkAvailableGames];
	
	if (!_availableGames.count) {
		[theMainWindow makeKeyAndOrderFront:self];
		[self locateGames];
	} else {
		[self finishSetup];
		[theMainWindow makeKeyAndOrderFront:self];
	}
}


- (void)finishSetup
{
	self.currentGame = [[NSUserDefaults standardUserDefaults] objectForKey:@"selectedGame"];
	if (!_currentGame || ![_availableGames containsObject:_currentGame]) {
		self.currentGame = [_availableGames firstObject];
	}
	
	startButton.hidden = NO;
	locateGameFilesButton.hidden = YES;
	locateGameFilesTextView.enclosingScrollView.hidden = YES;
	
	[self setupSmallCovers];
	
	mainCoverView.wantsLayer = YES;
	mainCoverViewOld.wantsLayer = YES;
	mainCoverViewOld_WA.wantsLayer = YES;
	animCoverView.wantsLayer = YES;

	mainCoverView.isMainCover = YES;
	mainCoverViewOld.isMainCover = YES;
	mainCoverViewOld_WA.isMainCover = YES;
	animCoverView.isMainCover = YES;

	mainCoverViewOld.hidden = YES;
	mainCoverViewOld_WA.hidden = YES;
	animCoverView.hidden = YES;
	animCoverView.layer.opacity = 0.0;
	
	mainCoverView.layer.zPosition = 10;
	animCoverView.layer.zPosition = 15;
	startButton.layer.zPosition = 15;
	musicButton.layer.zPosition = 15;
	
	[self setMainCover];
}


- (void)checkAvailableGames
{
	NSArray *games = @[@"pawn", @"guild", @"jinxter", @"corrupt", @"fish", @"myth", @"wonder", @"cguild", @"ccorrupt", @"cfish"];
	NSArray *names = @[@"The Pawn", @"The Guild of Thieves", @"Jinxter", @"Corruption", @"Fish!", @"Myth", @"Wonderland", @"The Guild of Thieves (MW)", @"Corruption (MW)", @"Fish! (MW)"];
	
	NSMutableArray *availableGames = [[[NSMutableArray alloc] init] autorelease];
	NSMutableDictionary *availableGamesData = [[[NSMutableDictionary alloc] init] autorelease];
	
	NSInteger i = 0;
	
	for (NSString *game in games) {
		
		if ([self checkForFile:[NSString stringWithFormat:@"%@.mag", game]]) {
			
			[availableGames addObject:game];
			
			NSMutableDictionary *data = [[[NSMutableDictionary alloc] init] autorelease];
			data[@"name"] = names[i];
			data[@"suffix"] = [NSString stringWithFormat:@"mX%li", i];
			data[@"cover"] = [NSImage imageNamed:[NSString stringWithFormat:@"cover_%@", game]];
			
			if (![self isCollection:game]) {
				data[@"cover_back"] = [NSImage imageNamed:[NSString stringWithFormat:@"cover_%@_back", game]];
			}
			
			NSString *musicName = game;
			
			if ([game isEqualToString:@"wonder"]) {
				data[@"hint"] = @"wonder";
			
			} else if ([self isCollection:game]) {
				data[@"hint"] = @"coll";
				
				if (!_coverCollection) {
					self.coverCollection = [NSImage imageNamed:@"cover_collection"];
					self.coverCollection_back = [NSImage imageNamed:@"cover_collection_back"];
				}
				
				data[@"altCover"] = _coverCollection;
				
				musicName = [game substringFromIndex:1];
				
			}
			
			NSString *musicPath = [self checkForFile:[NSString stringWithFormat:@"%@.mp3", musicName]];
			if (musicPath) {
				data[@"music"] = musicPath;
			}
			
			availableGamesData[game] = data;
			
		}
		
		i++;
		
	}
	
	
	if ([availableGames containsObject:@"cguild"]) { // navigation order!
		[availableGames removeObject:@"cguild"];
		[availableGames addObject:@"cguild"];
	}
	
	
	self.availableGames = availableGames;
	self.availableGamesData = availableGamesData;
	
	if (_availableGames.count) {
		[magneticController setIsAGameInstalled:YES];
	} else {
		[magneticController setIsAGameInstalled:NO];
	}
}


- (BOOL)isCollection:(NSString *)game
{
	if ([game isEqualToString:@"cguild"] || [game isEqualToString:@"ccorrupt"] || [game isEqualToString:@"cfish"]) {
		return YES;
	}
	
	return NO;
}


- (BOOL)isMagneticWindows:(NSString *)game
{
	if ([game isEqualToString:@"cguild"] || [game isEqualToString:@"ccorrupt"] || [game isEqualToString:@"cfish"] || [game isEqualToString:@"wonder"]) {
		return YES;
	}
	
	return NO;
}


- (BOOL)isCurrentGameCollection
{
	return [self isCollection:_currentGame];
}


- (BOOL)isCurrentGameMagneticWindows
{
	return [self isMagneticWindows:_currentGame];
}


- (void)setupSmallCovers
{
	if (!_currentGame) { return; }
	
	NSMutableArray *smallCoverViews = [[NSMutableArray alloc] init];
	
	BOOL usedAltCover = NO;
	
	for (NSString *game in _availableGames) {
		
		NSDictionary *gameData = _availableGamesData[game];
		
		NSImage *cover = gameData[@"altCover"];
		
		if (cover) {
			if (usedAltCover) { continue; } // collection only shows ONE cover here!
			usedAltCover = YES;
		
		} else {
			cover = gameData[@"cover"];
			
		}
		
		MXCoverView *iv = [[[MXCoverView alloc] initWithFrame:NSMakeRect(0, 0, 10, 10)] autorelease];
		iv.image = cover;
		[iv setWantsLayer:YES];
		iv.layer.opacity = 0.3;
		iv.layer.zPosition = 10;
		
		if (usedAltCover) {
			if (_availableGamesData[@"cfish"]) {
				iv.altTop = _availableGamesData[@"cfish"][@"cover"];
			}
			if (_availableGamesData[@"ccorrupt"]) {
				iv.altLeft = _availableGamesData[@"ccorrupt"][@"cover"];
			}
			if (_availableGamesData[@"cguild"]) {
				iv.altRight = _availableGamesData[@"cguild"][@"cover"];
			}
		}
		
		[selectSuperView addSubview:iv];
		
		[smallCoverViews addObject:iv];
		
	}
	
	self.smallCoverViews = [smallCoverViews autorelease];
}


- (NSInteger)currentGameIndex
{
	return [_availableGames indexOfObject:_currentGame];
}


- (void)updateCoverAnimType:(NSNotification *)notification
{
	[self refreshCoverAnimType:[notification.object integerValue]];
}


- (void)refreshCoverAnimType:(NSInteger)type
{
	self.coverAnimType = type;
	
	if (@available(macOS 14.0.0, *)) {
		
		if (type == 2) {
			mainCoverViewOldCurrent = mainCoverViewOld;
		} else {
			mainCoverViewOldCurrent = mainCoverViewOld_WA;
		}
		
	} else {
		mainCoverViewOldCurrent = mainCoverViewOld;
	}
}


- (void)turnCover
{	
	self.isBackCoverToggle = YES;

	if (!_showBackCover) {
		self.showBackCover = YES;
	} else {
		self.showBackCover = NO;
	}

	[self setMainCover];
	
	self.isBackCoverToggle = NO;
}


- (void)animCoverInDirection:(NSInteger)direction // 1 = from right, 2 = left, 3 = top
{
	NSInteger animType = self.coverAnimType;
	
	if (_isBackCoverToggle) { animType = 1; direction = 3; }
	
	NSRect bigCoverTarget = [self bigCoverTargetFrameFor:mainCoverView];
	NSRect bigCoverOldTarget = mainCoverViewOldCurrent.frame;
	NSRect bigCoverSource = NSZeroRect;
	
	if (animType == 1) {
		
		bigCoverSource = bigCoverTarget;
		CGFloat shadow = 30.0;
		
		if (direction == 1) {
			bigCoverSource.origin.x += (bigCoverSource.origin.x + bigCoverSource.size.width + shadow);
			bigCoverOldTarget.origin.x -= (bigCoverOldTarget.origin.x + bigCoverOldTarget.size.width + shadow);
		} else if (direction == 2) {
			bigCoverSource.origin.x -= (bigCoverSource.origin.x + bigCoverSource.size.width + shadow);
			bigCoverOldTarget.origin.x += (bigCoverOldTarget.origin.x + bigCoverOldTarget.size.width + shadow);
		} else if (direction == 3) {
			bigCoverSource.origin.y += (bigCoverSource.origin.y + bigCoverSource.size.height);
			bigCoverOldTarget.origin.y += (bigCoverOldTarget.origin.y + bigCoverOldTarget.size.height);
		}
		
	} else if (animType == 2) {
	
		bigCoverSource = newDeactivatedFrame;
		bigCoverOldTarget = oldDeactivatedFrame;
		
	}
	
	[CATransaction begin];
	[CATransaction setValue:@0.0f forKey:kCATransactionAnimationDuration];
	mainCoverView.isInAnim = YES;
	mainCoverViewOldCurrent.isInAnim = YES;
	mainCoverView.frame = bigCoverSource;
	mainCoverView.layer.position = bigCoverSource.origin;
	animCoverView.frame = bigCoverTarget;
	mainCoverViewOldCurrent.hidden = NO;
	mainCoverViewOldCurrent.layer.opacity = 1.0;
	[CATransaction commit];
	
	
	CGFloat duration = 0.3f;
	CAMediaTimingFunctionName functionName = nil;
	
	if (animType == 1) {
		duration = 0.3f;
		functionName = kCAMediaTimingFunctionEaseOut;
	} else {
		duration = 0.2f;
		functionName = kCAMediaTimingFunctionEaseIn;
	}
	
	NSRect old = mainCoverView.frame;
	NSRect new = bigCoverTarget, new2 = NSZeroRect;

	
	if (@available(macOS 14.0.0, *)) {
		if (animType == 2) {
			NSRect old2 = mainCoverViewOldCurrent.frame;
			new2 = bigCoverOldTarget;
			
			[selectSuperView setIsZooming:YES];

			WA_4SonomaAnimation *animation = [[[WA_4SonomaAnimation alloc] initWithView:mainCoverView startFrame:old endFrame:new view2:mainCoverViewOldCurrent startFrame2:old2 endFrame2:new2 completionBlock:^{
				
				[selectSuperView setIsZooming:NO];
				[self animationDidStop:nil finished:YES];
				
			}] autorelease];
			
			animation.animationCurve = NSAnimationEaseIn;
			animation.duration = duration;
			
			[animation startAnimation];
			
			return;
		}
	}
	
	
	NSRect oldBound = old;
	oldBound.origin = NSZeroPoint;
	
	NSRect newBound = new, newBound2 = NSZeroRect;
	newBound.origin = NSZeroPoint;
    
	[CATransaction begin];
	[CATransaction setValue:@(duration) forKey:kCATransactionAnimationDuration];
	
    CABasicAnimation *boundsAnimation = [CABasicAnimation animationWithKeyPath:@"bounds"];
    boundsAnimation.fromValue = [NSValue valueWithRect:oldBound];
    boundsAnimation.toValue = [NSValue valueWithRect:newBound];
    boundsAnimation.timingFunction = [CAMediaTimingFunction functionWithName:functionName];
	boundsAnimation.duration = duration;
	boundsAnimation.delegate = (id)self;
	[mainCoverView.layer addAnimation:boundsAnimation forKey:@"boundsAnimation"];
    
    CABasicAnimation *positionAnimation = [CABasicAnimation animationWithKeyPath:@"position"];
	positionAnimation.fromValue = [NSValue valueWithPoint:old.origin];
	positionAnimation.toValue = [NSValue valueWithPoint:new.origin];
    positionAnimation.timingFunction = [CAMediaTimingFunction functionWithName:functionName];
	positionAnimation.duration = duration;
    [mainCoverView.layer addAnimation:positionAnimation forKey:@"positionAnimation"];
    
	if (!mainCoverViewOldCurrent.hidden) {
		
		NSRect old2 = mainCoverViewOldCurrent.frame;
		new2 = bigCoverOldTarget;
		
		NSRect oldBound2 = old2;
		oldBound2.origin = NSZeroPoint;
		
		newBound2 = new2;
		newBound2.origin = NSZeroPoint;
		
		CABasicAnimation *boundsAnimation2 = [CABasicAnimation animationWithKeyPath:@"bounds"];
		boundsAnimation2.fromValue = [NSValue valueWithRect:oldBound2];
		boundsAnimation2.toValue = [NSValue valueWithRect:newBound2];
		boundsAnimation2.timingFunction = [CAMediaTimingFunction functionWithName:functionName];
		boundsAnimation2.duration = duration;
		[mainCoverViewOldCurrent.layer addAnimation:boundsAnimation2 forKey:@"boundsAnimation2"];
		
		CABasicAnimation *positionAnimation2 = [CABasicAnimation animationWithKeyPath:@"position"];
		positionAnimation2.fromValue = [NSValue valueWithPoint:old2.origin];
		positionAnimation2.toValue = [NSValue valueWithPoint:new2.origin];
		positionAnimation2.timingFunction = [CAMediaTimingFunction functionWithName:functionName];
		positionAnimation2.duration = duration;
		[mainCoverViewOldCurrent.layer addAnimation:positionAnimation2 forKey:@"positionAnimation2"];
		
		if (animType == 2) {
			CABasicAnimation *opacityAnimation = [CABasicAnimation animationWithKeyPath:@"opacity"];
			opacityAnimation.fromValue = @(1);
			opacityAnimation.toValue = @(0.3);
			opacityAnimation.timingFunction = [CAMediaTimingFunction functionWithName:functionName];
			opacityAnimation.duration = duration;
			[mainCoverViewOldCurrent.layer addAnimation:opacityAnimation forKey:@"opacityAnimation"];
		}
	}
	
	[CATransaction commit];

	[CATransaction begin];
	[CATransaction setValue:@0.0f forKey:kCATransactionAnimationDuration];
	
	mainCoverView.frame = new;
	[mainCoverView.layer setPosition:new.origin];
	[mainCoverView.layer setBounds:newBound];
	
	if (!mainCoverViewOldCurrent.hidden) {
		mainCoverViewOldCurrent.frame = new2;
		[mainCoverViewOldCurrent.layer setPosition:new2.origin];
		[mainCoverViewOldCurrent.layer setBounds:newBound2];
		
		if (animType == 2) {
			mainCoverViewOldCurrent.layer.opacity = 0.3;
		}
	}
	
	[CATransaction commit];
}


- (void)animationDidStop:(CAAnimation *)theAnimation finished:(BOOL)flag
{
	if (!flag) { return; }
	
	mainCoverView.isInAnim = NO;
	[mainCoverView setNeedsDisplay:YES];
	
	mainCoverViewOldCurrent.hidden = YES;
	
	if (self.coverAnimType == 2 && oldDeactivatedIndex >= 0) {
		
		MXCoverView *oldCV = _smallCoverViews[oldDeactivatedIndex];
		oldCV.enabled = YES;
		oldCV.layer.opacity = 0.3;
		
		NSShadow *smallShadow = [[[NSShadow alloc] init] autorelease];
		[smallShadow setShadowOffset:NSMakeSize(5.0, -5.0)];
		[smallShadow setShadowBlurRadius:6];
		[smallShadow setShadowColor:[NSColor colorWithCalibratedWhite:0.0 alpha:0.75]];
		oldCV.shadow = smallShadow;
		[oldCV setNeedsDisplay:YES];
		
		NSPoint windowLocation = [theMainWindow convertPointFromScreen:NSEvent.mouseLocation];
		NSPoint localPoint = [oldCV convertPoint:windowLocation fromView:nil];
		if (NSPointInRect(localPoint, oldCV.bounds)) {
			[oldCV activateWithMouseInsideAt:windowLocation];
		}
		
	}
	
	[self startCollectionCoverFading];
}


- (void)startCollectionCoverFading
{
	if ([self isCurrentGameCollection] && !self.showBackCover) {
		
		if (!animCoverView.image) {
			animCoverView.image = self.coverCollection;
		}
				
		animCoverView.hidden = NO;
		
		CABasicAnimation *blink = [CABasicAnimation animationWithKeyPath:@"opacity"];
		blink.fromValue = [NSNumber numberWithFloat:0.0];
		blink.toValue = [NSNumber numberWithFloat:1.0];
		blink.duration = 1;
		blink.autoreverses = YES;
		blink.repeatCount = FLT_MAX;
		blink.timingFunction = [CAMediaTimingFunction functionWithName:kCAMediaTimingFunctionEaseIn];
		
		[animCoverView.layer addAnimation:blink forKey:@"blinkAnim"];
	}
}


- (void)setMainCover
{
	if (!_availableGames.count) { return; }
	
	[animCoverView.layer removeAllAnimations];
	[mainCoverViewOldCurrent.layer removeAllAnimations];
	
	newDeactivatedFrame = NSZeroRect;
	oldDeactivatedFrame = NSZeroRect;
	oldDeactivatedIndex = -1;

	animCoverView.hidden = YES;
	[animCoverView.layer removeAllAnimations];
	
	BOOL isCollection = [self isCollection:_currentGame];
	NSImage *newImage = nil;
	
	if (!_showBackCover || !_isBackCoverToggle) {
		newImage = _availableGamesData[_currentGame][@"cover"];
		self.showBackCover = NO;
	} else {
		if (!isCollection) {
			newImage = _availableGamesData[_currentGame][@"cover_back"];
		} else {
			newImage = _coverCollection_back;
		}
		self.showBackCover = YES;
	}
	
	if (isCollection) {
		if (_showBackCover) {
			mainCoverView.collectionMainType = 2;
		} else {
			animCoverView.collectionMainType = 1;
			mainCoverView.collectionMainType = 0;
		}
	} else {
		mainCoverView.collectionMainType = 0;
	}
	
	
	BOOL didChange = NO;
	
	if (mainCoverView.image != newImage) {
		
		mainCoverViewOldCurrent.image = mainCoverView.image;
		mainCoverViewOldCurrent.frame = [self bigCoverTargetFrameFor:mainCoverView];
		
		mainCoverViewOldCurrent.layer.position = mainCoverViewOldCurrent.frame.origin;
		
		[mainCoverViewOldCurrent display]; // (!)
		
		for (MXCoverView *cv in _smallCoverViews) {
			if (!cv.isEnabled) {
				oldDeactivatedFrame = cv.frame;
				oldDeactivatedIndex = [_smallCoverViews indexOfObject:cv];
				break;
			}
		}
		
		mainCoverView.image = newImage;
		didChange = YES;
	}
	
	if (![_lastGame isEqualToString:_currentGame] && !([_lastGame hasSuffix:@"corrupt"] && [_currentGame hasSuffix:@"corrupt"]) && !([_lastGame hasSuffix:@"fish"] && [_currentGame hasSuffix:@"fish"])) {
		[self stopAllSounds];
	} else {
		didChange = NO;
	}
	
	NSShadow *smallShadow = [[[NSShadow alloc] init] autorelease];
	[smallShadow setShadowOffset:NSMakeSize(5.0, -5.0)];
	[smallShadow setShadowBlurRadius:6];
	[smallShadow setShadowColor:[NSColor colorWithCalibratedWhite:0.0 alpha:0.75]];
	
	for (NSImageView *iv in _smallCoverViews) {
		iv.enabled = YES;
		iv.shadow = smallShadow;
	}
	
	NSInteger disableIndex = self.currentGameIndex;
	if (self.isCurrentGameCollection) {
		disableIndex = _smallCoverViews.count - 1;
	}
	
	MXCoverView *currentCV = _smallCoverViews[disableIndex];
	currentCV.enabled = NO;
	currentCV.layer.opacity = 0.3;
	currentCV.shadow = nil;
	
	if (oldDeactivatedIndex >= 0 && _coverAnimType == 2) {
		MXCoverView *oldCV = _smallCoverViews[oldDeactivatedIndex];
		oldCV.enabled = NO;
		oldCV.layer.opacity = 0.3;
		oldCV.shadow = nil;
	}
	
	if (_availableGames.count == 1) {
		currentCV.type = 0;
	} else if (
			   
		([_currentGame isEqualToString:@"guild"] && [_availableGames containsObject:@"cguild"]) ||
		([_currentGame isEqualToString:@"corrupt"] && [_availableGames containsObject:@"ccorrupt"]) ||
		([_currentGame isEqualToString:@"fish"] && [_availableGames containsObject:@"cfish"]) ||
		
		([_currentGame isEqualToString:@"cguild"] && [_availableGames containsObject:@"guild"]) ||
		([_currentGame isEqualToString:@"ccorrupt"] && [_availableGames containsObject:@"corrupt"]) ||
		([_currentGame isEqualToString:@"cfish"] && [_availableGames containsObject:@"fish"])
		) {
		
		if (!_showBackCover) {
			currentCV.type = 2;
		} else {
			currentCV.type = 3;
		}
		
	} else {
		currentCV.type = 1;
	}
	
	[[NSUserDefaults standardUserDefaults] setObject:_currentGame forKey:@"selectedGame"];
	
	[self updateCoverViews];
	newDeactivatedFrame = currentCV.frame;
	
	NSInteger direction = 0;
	
	if (!self.coverAnimType || !_lastGame) {
		
		mainCoverView.frame = [self bigCoverTargetFrameFor:mainCoverView];
		animCoverView.frame = mainCoverView.frame;
		
	} else {
		
		if (![self isCollection:_lastGame] || !isCollection || _isBackCoverToggle) {
			
			NSInteger indexOfLast = [_availableGames indexOfObject:_lastGame];
			NSInteger indexOfCurrent = [_availableGames indexOfObject:_currentGame];
			
			if (indexOfLast < indexOfCurrent) {
				direction = 1;
				if (_isLooping) {
					direction = 2;
				}
			} else {
				direction = 2;
				if (_isLooping) {
					direction = 1;
				}
			}
			
			if (_isCollectionToggle &&
				
				(([_lastGame isEqualToString:@"guild"] && [_currentGame isEqualToString:@"cguild"]) ||
				([_lastGame isEqualToString:@"cguild"] && [_currentGame isEqualToString:@"guild"]) ||
				
				([_lastGame isEqualToString:@"corrupt"] && [_currentGame isEqualToString:@"ccorrupt"]) ||
				([_lastGame isEqualToString:@"ccorrupt"] && [_currentGame isEqualToString:@"corrupt"]) ||
				
				([_lastGame isEqualToString:@"fish"] && [_currentGame isEqualToString:@"cfish"]) ||
				([_lastGame isEqualToString:@"cfish"] && [_currentGame isEqualToString:@"fish"]))
				
				) {
				
				direction = 3;
				
			}
			
		} else {
			mainCoverViewOldCurrent.hidden = YES;
		}
		
	}
	
	self.lastGame = self.currentGame;
	
	if (direction) {
		[self animCoverInDirection:direction];
	} else {
		[self startCollectionCoverFading];
	}
	
	if (_availableGamesData[_currentGame][@"music"]) {
		musicButton.hidden = NO;
		
		if (didChange) {
			[self autoPlayTitleMusic];
		}
		
	} else {
		musicButton.hidden = YES;
	}
}


- (void)clickedInSelectionWindow:(NSNotification *)note
{
	NSInteger i = 0;
	
	for (MXCoverView *cv in _smallCoverViews) {
		
		if (cv.layer.opacity == 1.0) {
			
			self.currentGame = _availableGames[i];
			
			if ([self isCollection:_currentGame]) {
				
				if (cv.altTop == cv.image) {
					if (![self selectGameIfAvailable:@"cfish"]) {
						[self setMainCover];
					}
					
					return;
				}
				
				if (cv.altLeft == cv.image) {
					if (![self selectGameIfAvailable:@"ccorrupt"]) {
						[self setMainCover];
					}
					
					return;
				}
				
				if ([self selectGameIfAvailable:@"cguild"]) {
					return;
				}
				
			}
			
			[self setMainCover];
			return;
			
		}
		
		i++;
		
	}
}


- (void)doubleClickedMainCover:(NSNotification *)notification
{
	if (!_lastCollectionClickTurn || [NSDate.date timeIntervalSinceDate:_lastCollectionClickTurn] > 1) {
		[self start:self];
	}
}


- (void)clickedMainCoverInArea:(NSNotification *)notification
{
	if ([self isCurrentGameCollection]) {
		
		NSString *gameString = notification.object;
		
		if (_showBackCover) {
			
			if ([_availableGames containsObject:gameString]) {
				self.currentGame = gameString;
				self.lastCollectionClickTurn = NSDate.date;
				[self turnCover];
				return;
			}
			
		} else if (!_lastCollectionClickTurn || [NSDate.date timeIntervalSinceDate:_lastCollectionClickTurn] > 1) {
			[self selectGameIfAvailable:gameString];
		}
		
	}
}


- (void)clickedArrow:(NSNotification *)notification
{
	NSInteger type = [notification.object integerValue];
	if (type == 1) {
		[self changeGameSelectionWithNavigation:NSLeftArrowFunctionKey];
	} else if (type == 2) {
		[self changeGameSelectionWithNavigation:NSRightArrowFunctionKey];
	} else if (type == 3) {
		[self changeGameSelectionWithNavigation:NSUpArrowFunctionKey];
	}
}


- (void)changeGameSelectionWithNavigation:(unichar)character
{
	if ([magneticController selectedGame]) { return; }
	
	if (_availableGames.count < 2 && (character == NSLeftArrowFunctionKey || character == NSRightArrowFunctionKey)) { return; }
	
	if (character == ' ') {
		
		[musicButton performClick:self];
		
	} else if (character == NSLeftArrowFunctionKey) {
		
		NSInteger index = [self currentGameIndex];
		
		index --;
		if (index < 0) {
			_isLooping = YES;
			index = _availableGames.count - 1;
		}
		
		self.currentGame = _availableGames[index];
		
		if (_showBackCover && [self isCollection:_currentGame] && [self isCollection:_lastGame]) {
			[self turnCover];
		} else {
			[self setMainCover];
		}
		
		_isLooping = NO;

		return;
		
	} else if (character == NSRightArrowFunctionKey) {
		
		NSInteger index = [self currentGameIndex];
		
		index ++;
		if (index >= _availableGames.count) {
			_isLooping = YES;
			index = 0;
		}
		
		self.currentGame = _availableGames[index];
		
		if (_showBackCover && [self isCollection:_currentGame] && [self isCollection:_lastGame]) {
			[self turnCover];
		} else {
			[self setMainCover];
		}
		
		_isLooping = NO;
		
		return;
		
	} else if (character == NSDownArrowFunctionKey) {
		
		[self turnCover];
		
	} else if (!_showBackCover) {
		
		_isCollectionToggle = YES;
		
		if ([_currentGame isEqualToString:@"guild"]) {
			[self selectGameIfAvailable:@"cguild"];
		} else if ([_currentGame isEqualToString:@"corrupt"]) {
			[self selectGameIfAvailable:@"ccorrupt"];
		} else if ([_currentGame isEqualToString:@"fish"]) {
			[self selectGameIfAvailable:@"cfish"];
		} else if ([_currentGame isEqualToString:@"cguild"]) {
			[self selectGameIfAvailable:@"guild"];
		} else if ([_currentGame isEqualToString:@"ccorrupt"]) {
			[self selectGameIfAvailable:@"corrupt"];
		} else if ([_currentGame isEqualToString:@"cfish"]) {
			[self selectGameIfAvailable:@"fish"];
		}
		
		_isCollectionToggle = NO;
		
	}
}


- (void)changeGameSelectionWithDirectJump:(unichar)character
{
	[[NSNotificationCenter defaultCenter] postNotificationName:@"updateCoverAnimType" object:@(2)];

	if (character == 'p') {
		if (![_currentGame isEqualToString:@"pawn"]) {
			[self selectGameIfAvailable:@"pawn"];
		}

	} else if (character == 'g') {
		if ([_currentGame isEqualToString:@"guild"]) {
			[self selectGameIfAvailable:@"cguild"];
			
		} else if ([_currentGame isEqualToString:@"cguild"]) {
			[self selectGameIfAvailable:@"guild"];
			
		} else if ([self isCollection:_currentGame]) {
			if (![self selectGameIfAvailable:@"cguild"]) {
				[self selectGameIfAvailable:@"guild"];
			}
				   
		} else {
		   if (![self selectGameIfAvailable:@"guild"]) {
			   [self selectGameIfAvailable:@"cguild"];
		   }
		}
		
	} else if (character == 'j') {
		if (![_currentGame isEqualToString:@"jinxter"]) {
			[self selectGameIfAvailable:@"jinxter"];
		}
		
	} else if (character == 'c') {
		if ([_currentGame isEqualToString:@"corrupt"]) {
			[self selectGameIfAvailable:@"ccorrupt"];
			
		} else if ([_currentGame isEqualToString:@"ccorrupt"]) {
			[self selectGameIfAvailable:@"corrupt"];
			
		} else if ([self isCollection:_currentGame]) {
			if (![self selectGameIfAvailable:@"ccorrupt"]) {
				[self selectGameIfAvailable:@"corrupt"];
			}
			
		} else {
			if (![self selectGameIfAvailable:@"corrupt"]) {
				[self selectGameIfAvailable:@"ccorrupt"];
			}
		}
		
	} else if (character == 'f') {
		if ([_currentGame isEqualToString:@"fish"]) {
			[self selectGameIfAvailable:@"cfish"];
			
		} else if ([_currentGame isEqualToString:@"cfish"]) {
			[self selectGameIfAvailable:@"fish"];
			
		} else if ([self isCollection:_currentGame]) {
			if (![self selectGameIfAvailable:@"cfish"]) {
				[self selectGameIfAvailable:@"fish"];
			}
				   
		} else {
			if (![self selectGameIfAvailable:@"fish"]) {
				[self selectGameIfAvailable:@"cfish"];
			}
		}
		
	} else if (character == 'm') {
		if (![_currentGame isEqualToString:@"myth"]) {
			[self selectGameIfAvailable:@"myth"];
		}
		
	} else if (character == 'w') {
		if (![_currentGame isEqualToString:@"wonder"]) {
			[self selectGameIfAvailable:@"wonder"];
		}
		
	}
}


- (BOOL)selectGameIfAvailable:(NSString *)gameString
{
	if ([_availableGames containsObject:gameString]) {
		self.currentGame = gameString;
		
		if (_showBackCover && [self isCollection:_currentGame] && [self isCollection:_lastGame]) {
			[self refreshCoverAnimType:1];
			[self turnCover];
		} else {
			[self setMainCover];
		}
		
		return YES;
	}
	
	return NO;
}


- (void)locateGames
{
	startButton.hidden = YES;
	musicButton.hidden = YES;
	locateGameFilesButton.hidden = NO;
	locateGameFilesTextView.enclosingScrollView.hidden = NO;
}


- (IBAction)locateGameFiles:(id)sender
{
	NSOpenPanel *resultPanel = NSOpenPanel.openPanel;
	resultPanel.canChooseFiles = YES;
	resultPanel.canChooseDirectories = YES;
	resultPanel.allowsMultipleSelection = YES;
	
	[resultPanel beginSheetModalForWindow:theMainWindow completionHandler:^(NSInteger result) {
		if (NSModalResponseOK == result) {
			NSMutableArray *paths = NSMutableArray.array;
			for (NSURL *url in resultPanel.URLs) {
				[paths addObject:url.path];
			}
			[self startInstallationOf:paths];
        }
    }];
}


- (void)processGamesFiles:(NSNotification *)note
{
	[self startInstallationOf:note.object];
}


- (BOOL)startInstallationOf:(NSArray *)paths
{
	if ( ! _isSetup) {
		[self checkAvailableGames];
	}
	
	locateGameFilesButton.enabled = NO;
	importPIContainer.hidden = NO;
	[importPI startAnimation:self];
	
	BOOL result = [self installFiles:paths];
	
	if (result) {
		[self refreshAfterInstall];
	}
	
	locateGameFilesButton.enabled = YES;
	importPIContainer.hidden = YES;
	[importPI stopAnimation:self];
	
	return result;
}


- (void)playTitleMusic:(NSURL *)theUrl
{
	[self stopAllSounds];
	musicButton.state = NSOnState;
	
	AVAudioPlayer *mp3 = [[[AVAudioPlayer alloc] initWithContentsOfURL:theUrl error:nil] autorelease];
	self.titleMusicPlayer = mp3;
	
	mp3.numberOfLoops = 0;
	mp3.delegate = (id)self;
	[mp3 play];
}


- (void)audioPlayerDidFinishPlaying:(AVAudioPlayer *)player successfully:(BOOL)flag
{
	if (flag) { [self stopAllSounds]; }
}


- (void)playMidiData:(NSData *)data
{
	if ([_currentlyPlayingMidiData isEqualToData:data]) { return; }
	
	[self stopAllSounds];
	
	NSError *error = nil;
	AVMIDIPlayer *midi = [[[AVMIDIPlayer alloc] initWithData:data soundBankURL: nil error:&error] autorelease];
	if (error) { return; }
	
	self.wonderlandMidiPlayer = midi;
	self.currentlyPlayingMidiData = data;

	[midi play: ^{
		[self stopAllSounds];
	}];
}


- (void)stopAllSounds
{
	[self stopAllSounds:NO];
}


- (void)stopAllSounds:(BOOL)isFullStop
{
	[self.autoPlayDelayer invalidate];
	self.autoPlayDelayer = nil;
	
	[_wonderlandMidiPlayer stop];
	
	if (isFullStop) {
		[_titleMusicPlayer stop];
		self.titleMusicPlayer = nil;
		
	} else {
		[_titleMusicPlayer pause];
	}

	self.wonderlandMidiPlayer = nil;
	self.currentlyPlayingMidiData = nil;
	
	musicButton.state = NSOffState;
}


- (BOOL)isSoundPlaying
{
	if (_titleMusicPlayer.isPlaying || _wonderlandMidiPlayer.isPlaying) { return YES; }
	return NO;
}


- (void)autoPlayTitleMusic
{
	if ( ! [NSUserDefaults.standardUserDefaults boolForKey:@"themeAutoplay"]) { return; }
	[self.autoPlayDelayer invalidate];
	musicButton.state = NSOnState;
	self.autoPlayDelayer = [NSTimer scheduledTimerWithTimeInterval:0.5 target:self selector:@selector(toggleMusic:) userInfo:nil repeats:NO];
}


- (IBAction)toggleMusic:(id)sender
{
	[self.autoPlayDelayer invalidate];
	
	if (!self.isSoundPlaying) {
		[self playTitleMusic:[NSURL fileURLWithPath:_availableGamesData[_currentGame][@"music"]]];
	} else {
		[self stopAllSounds];
	}
}


- (void)updateUIStyle
{
	if ([magneticController selectedGame]) {
		[self updateWindowStyle];
	}
}


- (void)updateWindowStyle
{
	if ([magneticController currentStyle] != NSScrollerKnobStyleLight) {
		topBorder.appearance = [NSAppearance appearanceNamed:NSAppearanceNameVibrantLight];
		bottomBorder.appearance = [NSAppearance appearanceNamed:NSAppearanceNameVibrantLight];
		
		theMainWindow.appearance = [NSAppearance appearanceNamed:NSAppearanceNameAqua];
		
	} else {
		topBorder.appearance = [NSAppearance appearanceNamed:NSAppearanceNameVibrantDark];
		bottomBorder.appearance = [NSAppearance appearanceNamed:NSAppearanceNameVibrantDark];
		
		if (@available(macOS 10.14.0, *)) {
			theMainWindow.appearance = [NSAppearance appearanceNamed:NSAppearanceNameDarkAqua];
		} else {
			//theMainWindow.appearance = [NSAppearance appearanceNamed:NSAppearanceNameAqua];
		}
	}
}


- (IBAction)start:(id)sender
{
	if ([magneticController selectedGame]) { // SHOULD not happen ... but for nine years is could ... so to be totally sure ... *cough*
		return;
	}
	
	[selectSuperView retain];
	[theMainWindow setContentView:gamesSuperView];
	
	[topBorder removeFromSuperview];
	[gamesSuperView addSubview:topBorder];
	[bottomBorder removeFromSuperview];
	[gamesSuperView addSubview:bottomBorder];
	
	[self updateWindowStyle];
	
	[self stopAllSounds:YES];
	
	id currentlyVisibleGame = [self currentlyVisibleGame];
	
	[magneticController startGame:currentlyVisibleGame from:sender];
	
	theMainWindow.title = [NSString stringWithFormat:@"magnetiX - %@", currentlyVisibleGame[1]];
	NSApp.mainMenu.itemArray.firstObject.title = currentlyVisibleGame[1]; // *I* like it in THIS case!  :-)
	
	[theMainWindow makeKeyAndOrderFront:self];
	
	for (MXCoverView *smallView in _smallCoverViews) {
		[smallView endMouseOver];
	}
}


- (void)stop
{
	[gamesSuperView retain];
	[theMainWindow setContentView:selectSuperView];
	
	[topBorder removeFromSuperview];
	[selectSuperView addSubview:topBorder];
	[bottomBorder removeFromSuperview];
	[selectSuperView addSubview:bottomBorder];
	
	if (@available(macOS 10.14.0, *)) {
		theMainWindow.appearance = [NSAppearance appearanceNamed:NSAppearanceNameDarkAqua];
	} else {
		//theMainWindow.appearance = [NSAppearance appearanceNamed:NSAppearanceNameAqua];
	}
	topBorder.appearance = [NSAppearance appearanceNamed:NSAppearanceNameVibrantDark];
	bottomBorder.appearance = [NSAppearance appearanceNamed:NSAppearanceNameVibrantDark];
	
	theMainWindow.title = @"magnetiX - Please select game:";
	NSApp.mainMenu.itemArray.firstObject.title = @"magnetiX"; // *I* like it in THIS case!  :-)
	
	[[NSNotificationCenter defaultCenter] postNotificationName:@"updateCoverAnimType" object:@(0)];
	
	[self setMainCover];
	
	[self updateCoverViews];
	[selectSuperView repositionBackgroundLayer];
}


- (BOOL)validateMenuItem:(NSMenuItem *)menuItem
{
	if (menuItem.action == @selector(start:)) {
		if ([magneticController selectedGame] || !_availableGames.count || theMainWindow.sheets.count) {
			return NO;
		}
	} else if (menuItem.action == @selector(locateGameFiles:)) {
		if ([magneticController selectedGame] || theMainWindow.sheets.count) {
			return NO;
		}
	}
	
	return YES;
}


- (id)currentlyVisibleGame
{
	/*
	 theGames = [@[@[@"pawn", @"The Pawn", @"mX0"],
	 @[@"guild", @"The Guild of Thieves", @"mX1"],
	 @[@"jinxter" ,@"Jinxter", @"mX2"],
	 @[@"corrupt", @"Corruption", @"mX3"],
	 @[@"fish", @"Fish", @"mX4"],
	 @[@"myth", @"Myth", @"mX5"],
	 @[@"wonder", @"Wonderland", @"mX6", @"wonder"],
	 @[@"cguild", @"The Guild of Thieves (MW)", @"mX7", @"coll"],
	 @[@"ccorrupt", @"Corruption (MW)", @"mX8", @"coll"],
	 @[@"cfish", @"Fish (MW)", @"mX9", @"coll"]] retain];
	 */
	
	NSDictionary *data = _availableGamesData[_currentGame];
	
	NSMutableArray *dataInOldFormat = [[[NSMutableArray alloc] init] autorelease];
	[dataInOldFormat addObject:_currentGame];
	[dataInOldFormat addObject:data[@"name"]];
	[dataInOldFormat addObject:data[@"suffix"]];
	
	if (data[@"hint"]) {
		[dataInOldFormat addObject:data[@"hint"]];
	}
	
	return dataInOldFormat;
}


- (void)externalStart:(NSString *)filename
{
	int gameToSelect = 0;
	NSString *gameExtension = [filename pathExtension];
	
	if ([[[gameExtension substringToIndex:2] lowercaseString] isEqual:@"mx"]) {
		gameToSelect = [[gameExtension substringFromIndex:2] intValue];
		
		NSArray *games = @[@"pawn", @"guild", @"jinxter", @"corrupt", @"fish", @"myth", @"wonder", @"cguild", @"ccorrupt", @"cfish"];
		if ([_availableGames containsObject:games[gameToSelect]]) {
			self.currentGame = games[gameToSelect];
		} else {
			return;
		}
		
	} else { // scripts cannot be correctly assigned -> "normal start"
		[magneticController setExternalLoad:nil];
		return;
	}
	
	if ([self isCurrentGameMagneticWindows]) {
		[magneticController loadMagneticWindows];
	} else {
		[magneticController loadNONMagneticWindowsWithPath:filename];
	}
	
	[magneticController setIsLoadAtGameStart:YES];
	[self start:nil];
}


- (IBAction)openHelpUrl:(id)sender
{
	[[NSWorkspace sharedWorkspace] openURL:[NSURL URLWithString:@"http://www.maczentrisch.de/magnetiX/help.php"]];
}


- (BOOL)windowShouldClose:(id)sender
{
	if ([magneticController selectedGame]) {
	
		return [magneticController windowShouldClose:sender];
		
	}
	
	return YES;
}


- (void)windowDidBecomeMain:(NSNotification *)aNotification
{
	if ([windowMenu indexOfItemWithTag:666] != -1) {
		[windowMenu removeItemAtIndex:[windowMenu indexOfItemWithTag:666]];
	}
}


- (void)windowDidBecomeKey:(NSNotification *)aNotification
{
	NSFontManager *fontManager = [NSFontManager sharedFontManager];
	[[fontManager fontPanel:NO] orderOut:nil];
}


- (void)windowDidResize:(NSNotification *)aNotification
{
	if ([magneticController selectedGame]) {
	
	} else {
		
		[self updateCoverViews];
		[selectSuperView repositionBackgroundLayer];
		
	}
}


- (void)windowDidChangeScreen:(NSNotification *)notification
{
	[selectSuperView setupBackgroundLayer];
}


- (void)windowWillStartLiveResize:(NSNotification *)notification
{
	if ([magneticController selectedGame]) {
		[magneticController windowWillStartLiveResize:notification];
	}
}


- (void)windowDidEndLiveResize:(NSNotification *)notification
{
	if ([magneticController selectedGame]) {
		[magneticController windowDidEndLiveResize:notification];
	}
}


- (void)windowWillEnterFullScreen:(NSNotification *)notification
{
	[self deactivateAllCovers];
	[magneticController windowWillEnterFullScreen:notification];
	topBorder.hidden = YES;
	bottomBorder.hidden = YES;
}


- (void)windowWillExitFullScreen:(NSNotification *)notification
{
	[self deactivateAllCovers];
	[magneticController windowWillExitFullScreen:notification];
	topBorder.hidden = NO;
	bottomBorder.hidden = NO;
}



- (NSRect)window:(NSWindow *)window willPositionSheet:(NSWindow *)sheet usingRect:(NSRect)rect
{
	[self deactivateAllCovers];
	rect.origin.y -= 23;
	return rect;
}


- (NSRect)bigCoverTargetFrameFor:(NSImageView *)bigCoverView
{
	NSRect superFrame = bigCoverView.superview.frame;
	
	if (!topBorder.isHidden) { superFrame = NSInsetRect(superFrame, 0, 22); }
	
	NSRect targetFrame = superFrame;
	targetFrame.size.height = ceil(targetFrame.size.height / 100 * 80);
	targetFrame.origin.y += superFrame.size.height - targetFrame.size.height;
	
	targetFrame.size.height -= 44;
	
	// maxSize (wonderland) = 1096 * 1376
	
	CGFloat w = targetFrame.size.width / 1326 * 0.9 / 2;
	CGFloat h = targetFrame.size.height / 1376 * 0.9 / 2;
	
	if (w > h) { w = h; }
	
	NSSize imageSize = bigCoverView.image.size;
	
	imageSize.width = ceil(imageSize.width * w);
	imageSize.height = ceil(imageSize.height * w);
	
	NSRect frame = targetFrame;
	frame.size = imageSize;
	
	frame.origin.x = ceil(targetFrame.size.width / 2 - frame.size.width / 2);
	frame.origin.y += ceil(targetFrame.size.height / 2 - frame.size.height / 2);
	
	return frame;
}


- (void)updateCoverViews
{
	NSRect superFrame = mainCoverView.superview.frame;
	
	if (!topBorder.isHidden) { superFrame = NSInsetRect(superFrame, 0, 22); }

	
	NSRect targetFrame = superFrame;
	targetFrame.size.height = ceil(targetFrame.size.height / 100 * 80);
	targetFrame.origin.y += superFrame.size.height - targetFrame.size.height;
	
	targetFrame.size.height -= 44;
	
	// maxSize (wonderland) = 1096 * 1376
	
	CGFloat w = targetFrame.size.width / 1326 * 0.9 / 2;
	CGFloat h = targetFrame.size.height / 1376 * 0.9 / 2;
	
	if (w > h) { w = h; }
	
	NSSize imageSize = mainCoverView.image.size;
	
	imageSize.width = ceil(imageSize.width * w);
	imageSize.height = ceil(imageSize.height * w);
	
	NSRect frame = targetFrame;
	frame.size = imageSize;
	
	frame.origin.x = ceil(targetFrame.size.width / 2 - frame.size.width / 2);
	frame.origin.y += ceil(targetFrame.size.height / 2 - frame.size.height / 2);
	
	mainCoverView.frame = frame;
	animCoverView.frame = frame;
	
	// small one
	
	targetFrame = superFrame;
	targetFrame.size.height = ceil(targetFrame.size.height / 100 * 20);

	
	CGFloat borderLeft = 12;
	borderLeft += ceil((theMainWindow.frame.size.width - 880) / 2);
	if (borderLeft > 120) { borderLeft = 120; }
	
	CGFloat borderTop = 14;
	borderTop += ceil((theMainWindow.frame.size.height - 560) / 18);
	if (borderTop > 44) { borderTop = 44; }
	
	targetFrame = NSInsetRect(targetFrame, borderLeft, borderTop);
	targetFrame.size.width = targetFrame.size.width / 8; // _smallCoverViews.count;
	
	CGFloat centerCorr = ((8 - _smallCoverViews.count) * targetFrame.size.width) / 2;

	w = targetFrame.size.width / 1326 / 2;
	h = targetFrame.size.height / 1376 / 2;
	
	if (w > h) { w = h; }
	
	NSInteger i = 0;
	
	for (MXCoverView *smallView in _smallCoverViews) {
		
		NSSize imageSize = smallView.image.size;
		
		imageSize.width = ceil(imageSize.width * w);
		imageSize.height = ceil(imageSize.height * w);
		
		NSRect frame = targetFrame;
		frame.size = imageSize;
		
		frame.origin.x = targetFrame.origin.x + (targetFrame.size.width * i) + ((targetFrame.size.width - imageSize.width) / 2);
		frame.origin.x = ceil(frame.origin.x + centerCorr);
		
		frame.origin.y += ceil(targetFrame.size.height / 2 - frame.size.height / 2);
		
		smallView.frame = frame;
		
		[smallView removeTrackingArea:smallView.trackingArea];
		smallView.trackingArea = [[[NSTrackingArea alloc] initWithRect:NSInsetRect(smallView.bounds, 0, -20) options: (NSTrackingMouseEnteredAndExited | NSTrackingMouseMoved | NSTrackingActiveInKeyWindow) owner:smallView userInfo:nil] autorelease];
		[smallView addTrackingArea:smallView.trackingArea];
		
		i++;
	}
	
}


- (void)deactivateAllCovers
{
	if ([magneticController selectedGame]) { return; }
	
	for (MXCoverView *smallView in _smallCoverViews) {
		[smallView endMouseOver];
		if (smallView.overArrowType) {
			smallView.overArrowType = 0;
			[smallView setNeedsDisplay:YES];
		}
	}
}


- (id)magneticController
{
	return magneticController;
}


- (void)dealloc
{
	self.titleMusicPlayer = nil;
	self.wonderlandMidiPlayer = nil;
	self.currentlyPlayingMidiData = nil;
	self.altInstallPath = nil;
	self.availableGames = nil;
	self.availableGamesData = nil;
	self.coverCollection = nil;
	self.coverCollection_back = nil;
	self.currentGame = nil;
	self.lastGame = nil;
	self.sessionPath = nil;
	self.smallCoverViews = nil;
	self.lastCollectionClickTurn = nil;
	self.autoPlayDelayer = nil;
	
	[super dealloc];
}


@end
