/****************************************************************************\
 *
 * magnetiX - Magnetic Scrolls Interpreter - Mac OS X port
 *
 * Written by Jan-Sebastian Schliemann <magnetiX@maczentrisch.de>
 *
 *
 * the original Magnetic was:
 *
 * Written by Niclas Karlsson <nkarlsso@abo.fi>,
 *            David Kinder <davidk.kinder@virgin.net>,
 *            Stefan Meier <Stefan.Meier@if-legends.org> and
 *            Paul David Doherty <pdd@if-legends.org>
 *
 * Copyright (C) 1997-2008  Niclas Karlsson
 *
 *
 *
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.
 *
 \*****************************************************************************/

#import "MXCoverView.h"
#import "GSBackgroundView.h"

@implementation MXCoverView


- (NSDragOperation)draggingUpdated:(id <NSDraggingInfo>)sender
{
	return [self.window.contentView draggingUpdated:sender];
}

- (BOOL)prepareForDragOperation:(id <NSDraggingInfo>)sender
{
	return [self.window.contentView prepareForDragOperation:sender];
}

- (BOOL)performDragOperation:(id <NSDraggingInfo>)sender
{
	return [self.window.contentView performDragOperation:sender];
}


- (void)drawRect:(NSRect)rect
{
	BOOL isZooming = [(GSBackgroundView *)self.superview isZooming];
	
	if (isZooming) {
		[[NSGraphicsContext currentContext] setImageInterpolation:NSImageInterpolationNone];
	}
	
	[super drawRect:rect];

	if (!self.isEnabled) {
	
		NSRect rect = self.bounds;
		rect = NSInsetRect(rect, 0.5, 0.5);
		
		NSBezierPath *path = [NSBezierPath bezierPathWithRoundedRect:rect xRadius:3 yRadius:3];
		[[NSColor colorWithWhite:0.2 alpha:1.0] set];
		[path fill];
		[[NSColor colorWithWhite:0.6 alpha:1.0] set];
		[path stroke];
		
		
		[path removeAllPoints];
		[path moveToPoint:NSMakePoint(NSMidX(rect), NSMinY(rect) + 2.5)];
		[path lineToPoint:NSMakePoint(NSMidX(rect) + 10, NSMinY(rect) + 12.5)];
		[path lineToPoint:NSMakePoint(NSMidX(rect) - 10, NSMinY(rect) + 12.5)];
		[path closePath];
		
		if (_overArrowType == 4) {
			[[NSColor colorWithWhite:1 alpha:1] set];
		} else {
			[[NSColor colorWithWhite:0.6 alpha:0.75] set];
		}
		
		[path fill];
		
		
		if (_type < 1) { return; }
		
		CGFloat xtra = 0.0;
		if (_type < 2) {
			xtra = 6.0;
		}
		
		[path removeAllPoints];
		[path moveToPoint:NSMakePoint(NSMinX(rect) + 2.5, NSMidY(rect) + xtra)];
		[path lineToPoint:NSMakePoint(NSMinX(rect) + 12.5, NSMidY(rect) + 10 + xtra)];
		[path lineToPoint:NSMakePoint(NSMinX(rect) + 12.5, NSMidY(rect) - 10 + xtra)];
		[path closePath];
		
		if (_overArrowType == 1) {
			[[NSColor colorWithWhite:1 alpha:1] set];
		} else {
			[[NSColor colorWithWhite:0.6 alpha:0.75] set];
		}
		
		[path fill];
		
		[path removeAllPoints];
		[path moveToPoint:NSMakePoint(NSMaxX(rect) - 2.5, NSMidY(rect) + xtra)];
		[path lineToPoint:NSMakePoint(NSMaxX(rect) - 12.5, NSMidY(rect) + 10 + xtra)];
		[path lineToPoint:NSMakePoint(NSMaxX(rect) - 12.5, NSMidY(rect) - 10 + xtra)];
		[path closePath];
		
		if (_overArrowType == 2) {
			[[NSColor colorWithWhite:1 alpha:1] set];
		} else {
			[[NSColor colorWithWhite:0.6 alpha:0.75] set];
		}
		
		[path fill];
		
		if (_type < 2) { return; }
		
		[path removeAllPoints];
		[path moveToPoint:NSMakePoint(NSMidX(rect), NSMaxY(rect) - 2.5)];
		[path lineToPoint:NSMakePoint(NSMidX(rect) + 10, NSMaxY(rect) - 12.5)];
		[path lineToPoint:NSMakePoint(NSMidX(rect) - 10, NSMaxY(rect) - 12.5)];
		[path closePath];
		
		if (_type != 2) {
			[[NSColor colorWithWhite:0.6 alpha:0.3] set];
		} else if (_overArrowType == 3) {
			[[NSColor colorWithWhite:1 alpha:1] set];
		} else {
			[[NSColor colorWithWhite:0.6 alpha:0.75] set];
		}
		
		[path fill];
		
		return;
	}
	
	
	if (isZooming) {
	} else if (_isInAnim) {
		[[NSGraphicsContext currentContext] setImageInterpolation:NSImageInterpolationNone];
	} else if (self.inLiveResize) {
		[[NSGraphicsContext currentContext] setImageInterpolation:NSImageInterpolationMedium];
	} else {
		[[NSGraphicsContext currentContext] setImageInterpolation:NSImageInterpolationHigh];
	}
}


- (void)viewDidEndLiveResize
{
	[self setNeedsDisplay:YES];
}


- (void)mouseDown:(NSEvent *)event
{
	NSPoint event_location = event.locationInWindow;
	NSPoint local_point = [self convertPoint:event_location fromView:nil];
	
	if (!_isMainCover) {
		
		if (!self.isEnabled) {
			
			NSRect rect = self.bounds;
			rect = NSInsetRect(rect, 0.5, 0.5);
			
			CGFloat xtra = 0.0;
			if (_type < 2) {
				xtra = 6.0;
			}
			
			NSRect clickRect = NSMakeRect(0, 0, 20, 20);

			clickRect.origin = NSMakePoint(NSMidX(rect) - 10, NSMinY(rect));
			if (NSPointInRect(local_point, clickRect)) {
				[[NSNotificationCenter defaultCenter] postNotificationName:@"updateCoverAnimType" object:@(1)];
				[[NSNotificationCenter defaultCenter] postNotificationName:@"clickedTurn" object:nil];
				return;
			}
			
			if (_type) {
				
				clickRect.origin = NSMakePoint(NSMinX(rect), NSMidY(rect) - 10 + xtra);
				if (NSPointInRect(local_point, clickRect)) {
					[[NSNotificationCenter defaultCenter] postNotificationName:@"updateCoverAnimType" object:@(1)];
					[[NSNotificationCenter defaultCenter] postNotificationName:@"clickedArrow" object:@(1)];
					return;
				}
				
				clickRect.origin = NSMakePoint(NSMaxX(rect) - 20, NSMidY(rect) - 10 + xtra);
				if (NSPointInRect(local_point, clickRect)) {
					[[NSNotificationCenter defaultCenter] postNotificationName:@"updateCoverAnimType" object:@(1)];
					[[NSNotificationCenter defaultCenter] postNotificationName:@"clickedArrow" object:@(2)];
					return;
				}
				
				if (_type != 2) { return; }
				
				clickRect.origin = NSMakePoint(NSMidX(rect) - 10, NSMaxY(rect) - 20);
				if (NSPointInRect(local_point, clickRect)) {
					[[NSNotificationCenter defaultCenter] postNotificationName:@"updateCoverAnimType" object:@(1)];
					[[NSNotificationCenter defaultCenter] postNotificationName:@"clickedArrow" object:@(3)];
					return;
				}
				
			}
			
			return;
		}
		
		[[NSNotificationCenter defaultCenter] postNotificationName:@"updateCoverAnimType" object:@(2)];

		[super mouseDown:event];
		return;
	}
	
	if (!_collectionMainType) { return; }
	
	NSRect bounds = self.bounds;
	NSRect clickRect = NSZeroRect;
	
	if (_collectionMainType == 1) {
		
		NSBezierPath *pathCorrupt = [[[NSBezierPath alloc] init] autorelease];
			
		[pathCorrupt moveToPoint:NSMakePoint(bounds.size.width / 2, bounds.size.height / 100 * 18)];
		[pathCorrupt lineToPoint:NSMakePoint(bounds.size.width / 100 * 20, bounds.size.height / 100 * 18)];
		[pathCorrupt lineToPoint:NSMakePoint(bounds.size.width / 100 * 20, bounds.size.height / 100 * 50)];

		[pathCorrupt lineToPoint:NSMakePoint(bounds.size.width / 100 * 34, bounds.size.height / 100 * 60)];
		
		[pathCorrupt lineToPoint:NSMakePoint(bounds.size.width / 2, bounds.size.height / 100 * 50)];
		[pathCorrupt closePath];

		clickRect = NSMakeRect(bounds.size.width / 100 * 20, bounds.size.height / 100 * 7, bounds.size.width / 100 * 60, bounds.size.height / 100 * 4);

		if ([pathCorrupt containsPoint:local_point] || NSPointInRect(local_point, clickRect)) {
			[[NSNotificationCenter defaultCenter] postNotificationName:@"clickedMainCoverInArea" object:@"ccorrupt"];
			return;
		}
		
		
		NSBezierPath *pathGuild = [[[NSBezierPath alloc] init] autorelease];
			
		[pathGuild moveToPoint:NSMakePoint(bounds.size.width / 2, bounds.size.height / 100 * 18)];
		[pathGuild lineToPoint:NSMakePoint(bounds.size.width / 100 * 80, bounds.size.height / 100 * 18)];
		[pathGuild lineToPoint:NSMakePoint(bounds.size.width / 100 * 80, bounds.size.height / 100 * 50)];

		[pathGuild lineToPoint:NSMakePoint(bounds.size.width / 100 * 66, bounds.size.height / 100 * 60)];
		
		[pathGuild lineToPoint:NSMakePoint(bounds.size.width / 2, bounds.size.height / 100 * 50)];
		[pathGuild closePath];
		
		clickRect = NSMakeRect(bounds.size.width / 100 * 20, bounds.size.height / 100 * 3, bounds.size.width / 100 * 60, bounds.size.height / 100 * 4);

		
		if ([pathGuild containsPoint:local_point] || NSPointInRect(local_point, clickRect)) {
			[[NSNotificationCenter defaultCenter] postNotificationName:@"clickedMainCoverInArea" object:@"cguild"];
			return;
		}
		
		
		NSBezierPath *pathFish = [NSBezierPath bezierPathWithOvalInRect:NSMakeRect(bounds.size.width / 100 * 25, bounds.size.height / 100 * 56, bounds.size.width / 100 * 50, bounds.size.height / 100 * 30)];
		
		clickRect = NSMakeRect(bounds.size.width / 100 * 20, bounds.size.height / 100 * 11, bounds.size.width / 100 * 60, bounds.size.height / 100 * 4.5);
		
		if ([pathFish containsPoint:local_point] || NSPointInRect(local_point, clickRect)) {
			[[NSNotificationCenter defaultCenter] postNotificationName:@"clickedMainCoverInArea" object:@"cfish"];
			return;
		}
		
	} else {
		
		clickRect = NSMakeRect(bounds.size.width / 100 * 12, bounds.size.height / 100 * 20, bounds.size.width / 100 * 86, bounds.size.height / 100 * 8);
		if (NSPointInRect(local_point, clickRect)) {
			[[NSNotificationCenter defaultCenter] postNotificationName:@"clickedMainCoverInArea" object:@"ccorrupt"];
			return;
		}
		
		clickRect = NSMakeRect(bounds.size.width / 100 * 12, bounds.size.height / 100 * 10, bounds.size.width / 100 * 63, bounds.size.height / 100 * 10);
		if (NSPointInRect(local_point, clickRect)) {
			[[NSNotificationCenter defaultCenter] postNotificationName:@"clickedMainCoverInArea" object:@"cguild"];
			return;
		}
		
		clickRect = NSMakeRect(bounds.size.width / 100 * 2, bounds.size.height / 100 * 28, bounds.size.width / 100 * 96, bounds.size.height / 100 * 8);
		if (NSPointInRect(local_point, clickRect)) {
			[[NSNotificationCenter defaultCenter] postNotificationName:@"clickedMainCoverInArea" object:@"cfish"];
			return;
		}
		
	}
		
}


- (void)mouseUp:(NSEvent *)event
{
	if (_isMainCover && event.clickCount == 2) {
		[[NSNotificationCenter defaultCenter] postNotificationName:@"doubleClickedMainCover" object:nil];
	}
	[super mouseUp:event];
}


- (void)mouseEntered:(NSEvent *)event
{
	[self activateWithMouseInsideAt:event.locationInWindow];
}


- (void)activateWithMouseInsideAt:(NSPoint)locationInWindow
{
	self.isMouseOverNow = YES;
	
	if (self.isEnabled) {
		self.layer.opacity = 1.0;
		[(GSBackgroundView *)self.superview setMousedOverView:self];
	}
	
	[self mouseMovedToPoint:locationInWindow];
}


- (void)setEnabled:(BOOL)enabled
{
	[super setEnabled:enabled];
	
	if (!_isMouseOverNow) { return; }
	
	[self mouseMovedToPoint:[self.window convertPointFromScreen:NSEvent.mouseLocation]];
}


- (void)mouseMoved:(NSEvent *)event
{
	[self mouseMovedToPoint:event.locationInWindow];
}


- (void)mouseMovedToPoint:(NSPoint)locationInWindow
{
	if (!self.isEnabled) {
					
		NSPoint local_point = [self convertPoint:locationInWindow fromView:nil];
		
		NSRect rect = self.bounds;
		rect = NSInsetRect(rect, 0.5, 0.5);
		
		CGFloat xtra = 0.0;
		if (_type < 2) {
			xtra = 6.0;
		}
		
		NSInteger overArrowType = 0;
		
		NSRect clickRect = NSMakeRect(0, 0, 20, 20);

		clickRect.origin = NSMakePoint(NSMidX(rect) - 10, NSMinY(rect));
		if (NSPointInRect(local_point, clickRect)) {
			overArrowType = 4;
		}
		
		if (_type) {
			
			clickRect.origin = NSMakePoint(NSMinX(rect), NSMidY(rect) - 10 + xtra);
			if (NSPointInRect(local_point, clickRect)) {
				overArrowType = 1;
			}
			
			clickRect.origin = NSMakePoint(NSMaxX(rect) - 20, NSMidY(rect) - 10 + xtra);
			if (NSPointInRect(local_point, clickRect)) {
				overArrowType = 2;
			}
						
			clickRect.origin = NSMakePoint(NSMidX(rect) - 10, NSMaxY(rect) - 20);
			if (_type == 2 && NSPointInRect(local_point, clickRect)) {
				overArrowType = 3;
			}
			
		}
		
		if (overArrowType != _overArrowType) {
			_overArrowType = overArrowType;
			[self setNeedsDisplay:YES];
		}
		
		return;
		
	}

	if (_altTop || _altLeft || _altRight) {
		
		if (!_original) { self.original = self.image; }
		
		CGFloat extension = 100.0;
				
		NSRect frame = self.frame;
		
		NSRect testFrame = frame;
		testFrame.size.height = testFrame.size.height / 100 * 42;
		testFrame.origin.y += (frame.size.height - testFrame.size.height);
		
		testFrame.size.height += extension;
		testFrame = NSInsetRect(testFrame, -extension, 0);
		
		if (NSPointInRect(locationInWindow, testFrame)) {
			if (_altTop) {
				self.image = _altTop;
			} else if (_altLeft) {
				self.image = _altLeft;
			} else {
				self.image = _altRight;
			}
			
			return;
		}
		
		frame.size.height = frame.size.height / 100 * 58;
		
		testFrame = frame;
		testFrame.size.width = (testFrame.size.width / 2) + extension;
		testFrame.size.height += extension;
		testFrame.origin.x -= extension;
		testFrame.origin.y -= extension;
		
		if (NSPointInRect(locationInWindow, testFrame)) {
			if (_altLeft) {
				self.image = _altLeft;
			} else if (_altTop) {
				self.image = _altTop;
			} else {
				self.image = _altRight;
			}

			return;
		}

		if (_altRight) {
			self.image = _altRight;
		} else if (_altTop) {
			self.image = _altTop;
		} else {
			self.image = _altLeft;
		}

	}
		
}


- (void)mouseExited:(NSEvent *)event
{
	self.isMouseOverNow = NO;
	[self endMouseOver];
}


- (void)endMouseOver
{
	self.layer.opacity = 0.3;
	[(GSBackgroundView *)self.superview setMousedOverView:nil];
	
	if (_original) {
		self.image = self.original;
	}
	
	if (_overArrowType) {
		self.overArrowType = 0;
		[self setNeedsDisplay:YES];
	}
}


- (void)resizeWithOldSuperviewSize:(NSSize)oldSize
{

}


- (void)dealloc
{
	self.trackingArea = nil;
	self.altTop = nil;
	self.altLeft = nil;
	self.altRight = nil;
	self.original = nil;

	[super dealloc];
}


@end
