/****************************************************************************\
 *
 * magnetiX - Magnetic Scrolls Interpreter - Mac OS X port
 *
 * Written by Jan-Sebastian Schliemann <magnetiX@maczentrisch.de>
 *
 *
 * the original Magnetic was:
 *
 * Written by Niclas Karlsson <nkarlsso@abo.fi>,
 *            David Kinder <davidk.kinder@virgin.net>,
 *            Stefan Meier <Stefan.Meier@if-legends.org> and
 *            Paul David Doherty <pdd@if-legends.org>
 *
 * Copyright (C) 1997-2008  Niclas Karlsson
 *
 *
 *
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.
 *
 \*****************************************************************************/

#import "MXGameSelectWindow.h"
#import "GameSelectController.h"

@implementation MXGameSelectWindow


- (void)keyDown:(NSEvent *)theEvent
{
	NSString *characters = [theEvent characters];
	if (characters.length == 1 && !(theEvent.modifierFlags & NSEventModifierFlagCommand)) {
		unichar character = [characters characterAtIndex:0];
		if (character == NSLeftArrowFunctionKey || character == NSRightArrowFunctionKey || character == NSUpArrowFunctionKey || character == NSDownArrowFunctionKey || character == ' ') {
			[[NSNotificationCenter defaultCenter] postNotificationName:@"updateCoverAnimType" object:@(1)];
			[(GameSelectController *)[self delegate] changeGameSelectionWithNavigation:character];
			return;
		}
		
		if (character == 'p' || character == 'g' || character == 'j' || character == 'c' || character == 'f' || character == 'm' || character == 'w') {
			[(GameSelectController *)[self delegate] changeGameSelectionWithDirectJump:character];
			return;
		}
	}
	
	[super keyDown:theEvent];
}


- (BOOL)wantsScrollEventsForSwipeTrackingOnAxis:(NSEventGestureAxis)axis
{
	return YES;
}


- (void)scrollWheel:(NSEvent *)event
{
	if (event.phase == NSEventPhaseNone) { return; }
	if ( ! NSEvent.isSwipeTrackingFromScrollEventsEnabled) { return; }
	
	if (event.phase == NSEventPhaseBegan) {
		unichar character;
		if (fabs(event.scrollingDeltaY) <= fabs(event.scrollingDeltaX)) {
			if (event.scrollingDeltaX < 0) {
				character = NSRightArrowFunctionKey;
			} else {
				character = NSLeftArrowFunctionKey;
			}
		} else {
			if (event.scrollingDeltaY < 0) {
				character = NSDownArrowFunctionKey;
			} else {
				character = NSUpArrowFunctionKey;
			}
		}
		
		[[NSNotificationCenter defaultCenter] postNotificationName:@"updateCoverAnimType" object:@(1)];
		[(GameSelectController *)[self delegate] changeGameSelectionWithNavigation:character];
	}
}


@end
