/****************************************************************************\
*
* magnetiX - Magnetic Scrolls Interpreter - Mac OS X port
*
* Written by Jan-Sebastian Schliemann <magnetiX@maczentrisch.de>
*
*
* the original Magnetic was:
*
* Written by Niclas Karlsson <nkarlsso@abo.fi>,
*            David Kinder <davidk.kinder@virgin.net>,
*            Stefan Meier <Stefan.Meier@if-legends.org> and
*            Paul David Doherty <pdd@if-legends.org>
*
* Copyright (C) 1997-2008  Niclas Karlsson
*
*
*
*     This program is free software; you can redistribute it and/or modify
*     it under the terms of the GNU General Public License as published by
*     the Free Software Foundation; either version 2 of the License, or
*     (at your option) any later version.
*
*     This program is distributed in the hope that it will be useful,
*     but WITHOUT ANY WARRANTY; without even the implied warranty of
*     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*     GNU General Public License for more details.
*
*     You should have received a copy of the GNU General Public License
*     along with this program; if not, write to the Free Software
*     Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.
*
\*****************************************************************************/

#import "MXStatusView.h"

@implementation MXStatusView


- (void)drawRect:(NSRect)rect
{
	[super drawRect:rect];
	
	float lowerBorder = [[self window] contentBorderThicknessForEdge:NSMinYEdge];
	
	if (rect.origin.y < lowerBorder) {
		lowerBorder -= rect.origin.y;
		rect.origin.y += lowerBorder;
		rect.size.height -= lowerBorder;
	}
	
	[self.backgroundColor set];
	[NSBezierPath fillRect:rect];
}


- (void)mouseDown:(NSEvent *)event
{
	NSRect rect = self.frame;
	rect.size.height = 22;
	
	NSPoint event_location = event.locationInWindow;
	NSPoint local_point = [self convertPoint:event_location fromView:nil];
	
	if (NSPointInRect(local_point, rect)) {
		self.wasMouseDownInBorder = YES;
	} else {
		self.wasMouseDownInBorder = NO;
	}
}


- (void)mouseDragged:(NSEvent *)event
{
	if (self.wasMouseDownInBorder) {
		[self moveWindow:self.window withEvent:event];
	}

	[super mouseDragged:event];
}


- (void)moveWindow:(id)window withEvent:(NSEvent *)theEvent
{
	NSOperatingSystemVersion version = [[NSProcessInfo processInfo] operatingSystemVersion];
	if ((version.majorVersion == 10 && version.minorVersion >= 11) || version.majorVersion > 10) {
		
		[window performWindowDragWithEvent:theEvent];
		return;
		
	}
	
	NSPoint where = [theEvent locationInWindow];
	
	NSRect rect = NSMakeRect(1, 1, 1, 1);
	rect.origin = where;
	
	where =  [window convertRectToScreen:rect].origin;
	NSPoint origin = [window frame].origin;
	
	while ((theEvent = [NSApp nextEventMatchingMask:NSEventMaskLeftMouseDown|NSEventMaskLeftMouseDragged|NSEventMaskLeftMouseUp untilDate:[NSDate distantFuture] inMode:NSEventTrackingRunLoopMode dequeue:YES]) && ([theEvent type]!=NSEventTypeLeftMouseUp)) {
		NSAutoreleasePool  *pool = [[NSAutoreleasePool alloc] init];
		
		NSRect rect = NSMakeRect(1, 1, 1, 1);
		rect.origin = [theEvent locationInWindow];
		
		NSPoint now = [window convertRectToScreen:rect].origin;
		origin.x += now.x-where.x;
		origin.y += now.y-where.y;
		
		[window setFrameOrigin:origin];
		where = now;
		[pool release];
	}
}


- (void)scrollWheel:(NSEvent *)theEvent
{
	[visibleScrollView scrollWheel:theEvent];
}

- (void)resetCursorRects
{
	[super resetCursorRects];
	[self addCursorRect:self.visibleRect cursor:[NSCursor arrowCursor]];
}

- (void)dealloc
{
	self.backgroundColor = nil;
	
	[super dealloc];
}

@end
